#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import NOOP, Instrumentation, trace
from freighter import Payload, UnaryClient

from synnax.ranger.payload import RangePayload, RangeKey


class _CreateRequest(Payload):
    ranges: list[RangePayload]


_CreateResponse = _CreateRequest


class _DeleteRequest(Payload):
    keys: list[RangeKey]


class _DeleteResponse(Payload):
    ...


_CREATE_ENDPOINT = "/range/create"
_DELETE_ENDPOINT = "/range/delete"


class RangeWriter:
    __client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self.__client = client
        self.instrumentation = instrumentation

    @trace("debug", "range.create")
    def create(self, ranges: list[RangePayload]) -> list[RangePayload]:
        req = _CreateRequest(ranges=ranges)
        res, exc = self.__client.send(_CREATE_ENDPOINT, req, _CreateResponse)
        if exc is not None:
            raise exc
        return res.ranges

    @trace("debug", "range.delete")
    def delete(self, keys: list[RangeKey]):
        req = _DeleteRequest(keys=keys)
        res, exc = self.__client.send(_DELETE_ENDPOINT, req, _DeleteResponse)
        if exc is not None:
            raise exc



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.
import functools
from typing import overload, Callable

from freighter import UnaryClient

from synnax.channel.retrieve import ChannelRetriever
from synnax.exceptions import QueryError
from synnax.framer.client import Client
from synnax.ranger.alias import Aliaser
from synnax.ranger.writer import RangeWriter
from synnax.ranger.active import Active
from synnax.ranger.kv import KV
from synnax.ranger.payload import (
    RangeKey,
    RangeKeys,
    RangeName,
    RangeNames,
    RangeParams,
    RangePayload,
    normalize_range_params,
)
from synnax.ranger.range import Range
from synnax.ranger.retrieve import RangeRetriever
from synnax.telem import TimeRange
from synnax.signals.signals import Registry
from synnax.state import LatestState

RANGE_SET_CHANNEL = "sy_range_set"


class RangeClient:
    __frame_client: Client
    __channel_retriever: ChannelRetriever
    __retriever: RangeRetriever
    __writer: RangeWriter
    __unary_client: UnaryClient
    __active: Active
    __signals: Registry

    def __init__(
        self,
        unary_client: UnaryClient,
        frame_client: Client,
        writer: RangeWriter,
        retriever: RangeRetriever,
        channel_retriever: ChannelRetriever,
        signals: Registry,
    ) -> None:
        self.__unary_client = unary_client
        self.__frame_client = frame_client
        self.__writer = writer
        self.__retriever = retriever
        self.__channel_retriever = channel_retriever
        self.__active = Active(self.__unary_client)
        self.__signals = signals

    @overload
    def create(
        self,
        *,
        name: str,
        time_range: TimeRange,
        retrieve_if_name_exists: bool = False,
    ) -> Range:
        """Creates a named range spanning a region of time. This range is persisted
        to the cluster and is visible to all clients.

        :param name: The name of the range
        :param time_range: The time range of the range. This time range must be valid
        i.e. its start time must be less than or equal to its end time.
        :param retrieve_if_name_exists: If true, this method will retrieve the range
        if it already exists. This method will throw a ValidationError
        if the range already exists and has a DIFFERENT time range.
        """
        ...

    @overload
    def create(
        self,
        ranges: Range,
        retrieve_if_name_exists: bool = False,
    ) -> Range:
        """Creates the given range. This range is persisted to the cluster and is
        visible to all clients.

        :param ranges: The range to create
        :param retrieve_if_name_exists: If true, this method will retrieve the range
        if it already exists. This method will throw a ValidationError if the range
        already exists and has a DIFFERENT time range.
        """
        ...

    @overload
    def create(
        self,
        ranges: list[Range],
        retrieve_if_name_exists: bool = False,
    ) -> list[Range]:
        """Creates the given ranges. These ranges are persisted to the cluster and are
        visible to all clients.

        :param ranges: The ranges to create
        :param retrieve_if_name_exists: If true, this method will retrieve the range
        if it already exists. This method will throw a ValidationError if the range
        already exists and has a DIFFERENT time range.
        """
        ...

    def create(
        self,
        ranges: Range | list[Range] = None,
        *,
        name: str = "",
        time_range: TimeRange | None = None,
        retrieve_if_name_exists: bool = False,
    ) -> Range | list[Range]:
        is_single = True
        if ranges is None:
            to_create = [RangePayload(name=name, time_range=time_range)]
        elif isinstance(ranges, Range):
            to_create = [ranges.to_payload()]
        else:
            is_single = False
            to_create = [r.to_payload() for r in ranges]

        res: list[Range] = list()
        if retrieve_if_name_exists:
            res = self.retrieve([r.name for r in to_create])
            if is_single and len(res) > 1:
                filtered = [r for r in res if r.time_range == time_range]
                if len(filtered) == 0:
                    raise QueryError(
                        f"""
                        retrieve_if_name_exists was set to true, but {len(res)} ranges
                        were found matching {name} but none had the same time range as
                        passed to create. Synnax can't figure out which one you want!
                        """
                    )
                if len(filtered) > 1:
                    raise QueryError(
                        f"""
                    retrieve_if_name_exists was set to true, but {len(res)} ranges were
                    found that matched {name} and had time range {time_range}. Synnax
                    can't figure out which one you want!
                    """
                    )
                res = [filtered[0]]
            existing_names = {r.name for r in res}
            to_create = [r for r in to_create if r.name not in existing_names]

        if len(to_create) > 0:
            res.extend(self.__sugar(self.__writer.create(to_create)))

        return res if not is_single else res[0]

    @overload
    def retrieve(
        self,
        key_or_name: RangeKey | RangeName,
    ) -> Range:
        ...

    @overload
    def retrieve(
        self,
        keys_or_names: RangeKeys | RangeNames,
    ) -> list[Range]:
        ...

    def retrieve(
        self,
        params: RangeParams,
    ) -> Range | list[Range]:
        normal = normalize_range_params(params)
        _ranges = self.__retriever.retrieve(normal.params)
        sug = self.__sugar(_ranges)
        if not normal.single:
            return sug
        if len(sug) == 0:
            raise QueryError(f"Range {normal} not found")
        elif len(sug) > 1:
            raise QueryError(f"Multiple ranges matching {normal} found")
        return sug[0]

    def delete(
        self,
        params: RangeParams,
    ) -> None:
        _ranges = self.__retriever.retrieve(params)
        self.__writer.delete([r.key for r in _ranges])

    def set_active(self, key: RangeKey):
        self.__active.set(key)

    def clear_active(self):
        self.__active.clear()

    def retrieve_active(self) -> Range | None:
        rng = self.__active.retrieve()
        return None if rng is None else self.__sugar([rng])[0]

    def search(
        self,
        term: str,
    ) -> list[Range]:
        _ranges = self.__retriever.search(term)
        return self.__sugar(_ranges)

    def __sugar(self, ranges: list[RangePayload]):
        return [
            Range(
                **r.dict(),
                _frame_client=self.__frame_client,
                _channel_retriever=self.__channel_retriever,
                _kv=KV(r.key, self.__unary_client),
                _aliaser=Aliaser(r.key, self.__unary_client),
            )
            for r in ranges
        ]

    def on_create(self, f: Callable[[Range], None]):
        @functools.wraps(f)
        def wrapper(state: LatestState):
            d = state[RANGE_SET_CHANNEL]
            f(
                Range(
                    name=d["name"],
                    key=d["key"],
                    time_range=TimeRange(
                        start=d["time_range"]["start"],
                        end=d["time_range"]["end"],
                    ),
                    _frame_client=self.__frame_client,
                    _channel_retriever=self.__channel_retriever,
                    _kv=KV(d["key"], self.__unary_client),
                    _aliaser=Aliaser(d["key"], self.__unary_client),
                )
            )

        return self.__signals.on([RANGE_SET_CHANNEL], lambda s: True)(wrapper)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

import uuid
from asyncio import create_task, events, tasks
from collections.abc import Callable
from threading import Event, Thread
from typing import Any, Protocol

import numpy as np
from janus import Queue

from synnax import framer
from synnax.channel.payload import (
    ChannelKey,
    ChannelName,
    ChannelParams,
    ChannelPayload,
)
from synnax.channel.retrieve import ChannelRetriever, retrieve_required
from synnax.telem import CrudeTimeSpan, TimeSpan, TimeStamp
from synnax.telem.control import Authority, CrudeAuthority


class State:
    value: dict[ChannelKey, np.number]
    __retriever: ChannelRetriever

    def __init__(self, retrieve: ChannelRetriever):
        self.__retriever = retrieve
        self.value = dict()

    def update(self, value: framer.Frame):
        for i, key in enumerate(value.columns):
            self.value[key] = value.series[i][0]

    def __getattr__(self, ch: ChannelKey | ChannelName | ChannelPayload):
        ch = retrieve_required(self.__retriever, ch)[0]
        return self.value[ch.key]

    def __getitem__(self, ch: ChannelKey | ChannelName | ChannelPayload):
        return self.__getattr__(ch)


class Processor(Protocol):
    def process(self, state: Controller) -> Any:
        ...


class WaitUntil(Processor):
    event: Event
    callback: Callable[[Controller], bool]

    def __init__(self, callback: Callable[[Controller], bool]):
        self.event = Event()
        self.callback = callback

    def process(self, state: Controller) -> Any:
        if self.callback(state):
            self.event.set()
        return None


class Controller:
    writer: framer.Writer
    idx_map: dict[ChannelKey, ChannelKey]
    retriever: ChannelRetriever
    receiver: _Receiver

    def __init__(
        self,
        name: str,
        write: ChannelParams,
        read: ChannelParams,
        frame_client: framer.Client,
        retriever: ChannelRetriever,
        write_authorities: CrudeAuthority | list[CrudeAuthority],
    ) -> None:
        write_channels = retrieve_required(retriever, write)
        write_keys = [ch.index for ch in write_channels if ch.index != 0]
        write_keys.extend([ch.key for ch in write_channels])
        self.writer = frame_client.new_writer(
            name=name,
            start=TimeStamp.now(),
            channels=write_keys,
            authorities=write_authorities,
        )
        self.receiver = _Receiver(frame_client, read, retriever, self)
        self.retriever = retriever
        self.receiver.start()
        self.receiver.bootup_ack.wait()

    def set(
        self,
        ch: ChannelKey | ChannelName | dict[ChannelKey | ChannelName, int | float],
        value: int | float | None = None,
    ):
        if isinstance(ch, dict):
            values = list(ch.values())
            channels = retrieve_required(self.retriever, list(ch.keys()))
            now = TimeStamp.now()
            updated = {channels[i].key: values[i] for i in range(len(channels))}
            updated_idx = {channels[i].index: now for i in range(len(channels))}
            self.writer.write({**updated, **updated_idx})
            return
        ch = retrieve_required(self.retriever, ch)[0]
        self.writer.write({ch.key: value, ch.index: TimeStamp.now()})

    def authorize(self, ch: ChannelKey | ChannelName, value: Authority):
        ch = retrieve_required(self.retriever, ch)[0]
        self.writer.set_authority({ch.key: value, ch.index: value})

    def wait_until(
        self,
        callback: Callable[[Controller], bool],
        timeout: CrudeTimeSpan = None,
    ) -> bool:
        processor = WaitUntil(callback)
        key = uuid.uuid4()
        try:
            self.receiver.processors[key] = processor
            ok = processor.event.wait(
                timeout=TimeSpan(timeout).seconds if timeout else None
            )
        finally:
            del self.receiver.processors[key]
        return ok

    def release(self):
        self.writer.close()
        self.receiver.close()

    def __setitem__(
        self, ch: ChannelKey | ChannelName | ChannelPayload, value: int | float
    ):
        self.set(ch, value)

    def __setattr__(self, key, value):
        try:
            super().__setattr__(key, value)
        except AttributeError:
            self.set(key, value)

    def get(self, ch: ChannelKey | ChannelName) -> int | float:
        ch = retrieve_required(self.retriever, ch)[0]
        return self.receiver.state[ch.key]

    def __getitem__(self, item):
        return self.get(item)

    def __getattr__(self, item):
        try:
            return super().__getattr__(item)
        except AttributeError:
            return self.get(item)

    def __enter__(self) -> Controller:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.release()


def cancel_all_tasks(loop):
    to_cancel = tasks.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(tasks.gather(*to_cancel, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


class _Receiver(Thread):
    state: State
    channels: ChannelParams
    client: framer.Client
    streamer: framer.AsyncStreamer
    processors: dict[uuid.UUID, Processor]
    retriever: ChannelRetriever
    controller: Controller
    bootup_ack: Event

    def __init__(
        self,
        client: framer.Client,
        channels: ChannelParams,
        retriever: ChannelRetriever,
        controller: Controller,
    ):
        self.channels = retriever.retrieve(channels)
        self.client = client
        self.state = State(retriever)
        self.controller = controller
        self.bootup_ack = Event()
        self.processors = {}
        super().__init__()

    def run(self):
        loop = events.new_event_loop()
        try:
            events.set_event_loop(loop)
            loop.run_until_complete(self.__run())
        finally:
            try:
                cancel_all_tasks(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.run_until_complete(loop.shutdown_default_executor())
            finally:
                events.set_event_loop(None)
                loop.close()

    def __process(self):
        for processor in self.processors.values():
            processor.process(self.controller)

    async def __listen_for_close(self):
        await self.queue.async_q.get()
        await self.streamer.close_loop()

    async def __run(self):
        self.queue = Queue(maxsize=1)
        self.streamer = await self.client.new_async_streamer(self.channels)
        self.bootup_ack.set()
        create_task(self.__listen_for_close())

        async for frame in self.streamer:
            self.state.update(frame)
            self.__process()

    def close(self):
        self.queue.sync_q.put(None)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from freighter import (
    EOF,
    AsyncStream,
    AsyncStreamClient,
    ExceptionPayload,
    Payload,
    Stream,
    StreamClient,
)

from synnax.channel.payload import ChannelKeys
from synnax.exceptions import UnexpectedError
from synnax.framer.adapter import ReadFrameAdapter
from synnax.framer.frame import Frame, FramePayload
from synnax.telem import CrudeTimeStamp, TimeStamp


class _Request(Payload):
    start: TimeStamp
    keys: ChannelKeys


class _Response(Payload):
    frame: FramePayload
    error: ExceptionPayload | None


_ENDPOINT = "/frame/stream"


class Streamer:
    __stream: Stream[_Request, _Response]
    __adapter: ReadFrameAdapter
    from_: CrudeTimeStamp

    def __init__(
        self,
        client: StreamClient,
        adapter: ReadFrameAdapter,
        from_: CrudeTimeStamp | None = None,
    ) -> None:
        self.from_ = from_ or TimeStamp.now()
        self.__stream = client.stream(_ENDPOINT, _Request, _Response)
        self.__adapter = adapter
        self.__open()

    def __open(self):
        self.__stream.send(_Request(keys=self.__adapter.keys, start=self.from_))

    @property
    def received(self) -> bool:
        return self.__stream.received()

    def read(self) -> Frame:
        res, err = self.__stream.receive()
        if err is not None:
            raise err
        return self.__adapter.adapt(Frame(res.frame))

    def close(self):
        exc = self.__stream.close_send()
        if exc is not None:
            raise exc
        _, exc = self.__stream.receive()
        if exc is None:
            raise UnexpectedError(
                """Unexpected missing close acknowledgement from server.
                Please report this issue to the Synnax team."""
            )
        elif not isinstance(exc, EOF):
            raise exc

    def __iter__(self):
        return self

    def __enter__(self):
        return self

    def __next__(self):
        return self.read()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class AsyncStreamer:
    __stream: AsyncStream[_Request, _Response]
    __client: AsyncStreamClient
    __adapter: ReadFrameAdapter
    from_: CrudeTimeStamp

    def __init__(
        self,
        client: AsyncStreamClient,
        adapter: ReadFrameAdapter,
        from_: CrudeTimeStamp | None = None,
    ) -> None:
        self.from_ = from_ or TimeStamp.now()
        self.__client = client
        self.__adapter = adapter

    async def open(self):
        self.__stream = await self.__client.stream(_ENDPOINT, _Request, _Response)
        await self.__stream.send(_Request(keys=self.__adapter.keys, start=self.from_))

    @property
    def received(self) -> bool:
        return self.__stream.received()

    async def read(self) -> Frame:
        res, err = await self.__stream.receive()
        if err is not None:
            raise err
        return self.__adapter.adapt(Frame(res.frame))

    async def close_loop(self):
        await self.__stream.close_send()

    async def close(self):
        exc = await self.__stream.close_send()
        if exc is not None:
            raise exc
        _, exc = await self.__stream.receive()
        if exc is None:
            raise UnexpectedError(
                """Unexpected missing close acknowledgement from server.
                Please report this issue to the Synnax team."""
            )
        elif not isinstance(exc, EOF):
            raise exc

    async def __aenter__(self):
        return self

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return await self.read()
        except EOF:
            raise StopAsyncIteration

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
